#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define N (1<<26)

#define timeit(id, fn) {\
  double res, start, end; \
  start = get_time(); \
  res = fn; \
  end = get_time(); \
  printf("%i %g %g\n", id, res, end-start); \
}


// get_time from https://stackoverflow.com/a/2349941
#ifdef WIN32
#include <windows.h>
double get_time()
{
    LARGE_INTEGER t, f;
    QueryPerformanceCounter(&t);
    QueryPerformanceFrequency(&f);
    return (double)t.QuadPart/(double)f.QuadPart;
}
#else
#include <sys/time.h>
#include <sys/resource.h>
double get_time()
{
    struct timeval t;
    struct timezone tzp;
    gettimeofday(&t, &tzp);
    return t.tv_sec + t.tv_usec*1e-6;
}
#endif


/*
Structs:
xy:     a simple struct that can be put in an array
xy_arr: a struct of arrays
*/
struct xy {
  double x, y;
  double other, stuff, because, that, happens;  
};
struct xy_arr {
  double *restrict x, *restrict y;
  double *other, *stuff, *because, *that, *happens;  
};


// Calculate sum of x + 2*y for an array of structs
double x_plus_fy_1( struct xy *s_xy ) {
  double sum = 0;
  for ( int i=0; i<N; i++ ) {
    sum += s_xy[i].x + 2 * s_xy[i].y;
  }
  return sum;
}

// Calculate sum of x + 2*y for the arrays in a struct
double x_plus_fy_2( struct xy_arr xy_a ) {
  double sum = 0;
  for ( int i=0; i<N; i++ ) {
    sum += xy_a.x[i] + 2 * xy_a.y[i];
  }
  return sum;
}

// Calculate sum of x + 2*y for the arrays in a struct, being even more local in accesses
double x_plus_fy_3( struct xy_arr xy_a ) {
  double sum = 0;
  for ( int i=0; i<N; i++ ) {
    sum += xy_a.x[i];
  }
  for ( int i=0; i<N; i++ ) {
    sum += 2 * xy_a.y[i];
  }
  return sum;
}

int main() {
  struct xy *s_xy = malloc(N * sizeof(struct xy));
  struct xy_arr xy_a;
  xy_a.x = malloc(N * sizeof(double));
  xy_a.y = malloc(N * sizeof(double));

  double x, y;
  srand(time(0));
  
  // initialize the arrays identically
  for ( int i=0; i<N; i++ ) {
    x = (double)rand()/RAND_MAX;
    y = (double)rand()/RAND_MAX;
    s_xy[i].x = x;
    s_xy[i].y = y;
    xy_a.x[i] = x;
    xy_a.y[i] = y;
  }
  
  // time the implementations
  timeit(1, x_plus_fy_1(s_xy));
  timeit(2, x_plus_fy_2(xy_a));
  timeit(3, x_plus_fy_3(xy_a));
  
  return 0;
}




